import React, { useState, useEffect } from 'react';
import PropTypes from 'prop-types';
import { transform } from 'sucrase';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Date.prototype.toString.call(Reflect.construct(Date, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _construct(Parent, args, Class) {
  if (isNativeReflectConstruct()) {
    _construct = Reflect.construct;
  } else {
    _construct = function _construct(Parent, args, Class) {
      var a = [null];
      a.push.apply(a, args);
      var Constructor = Function.bind.apply(Parent, a);
      var instance = new Constructor();
      if (Class) _setPrototypeOf(instance, Class.prototype);
      return instance;
    };
  }

  return _construct.apply(null, arguments);
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

var evalCode = function evalCode(code, scope) {
  var scopeKeys = Object.keys(scope);
  var scopeValues = Object.values(scope); // eslint-disable-next-line no-new-func

  var fn = _construct(Function, scopeKeys.concat([code]));

  return fn.apply(void 0, scopeValues);
};

var prepareCode = function prepareCode(code) {
  var exportRegexp = /\bexport default\b/gm;
  var renderRegexp = /\brender(\([^)])/gm;
  var elementRegexp = /^<.+>/g; // export default Component

  if (exportRegexp.test(code)) return code.replace(exportRegexp, 'return'); // render(<Component />)

  if (renderRegexp.test(code)) return code.replace(renderRegexp, 'return $1'); // remove trailing comma for expression

  code = code.replace(/;$/, ''); // inline elements

  if (elementRegexp.test(code) && React.Fragment) code = "<>" + code + "</>";
  return "return (" + code + ")";
};

var errorBoundary = function errorBoundary(Element, errorCallback) {
  var _temp;

  return _temp =
  /*#__PURE__*/
  function (_React$Component) {
    _inheritsLoose(ErrorBoundary, _React$Component);

    function ErrorBoundary() {
      var _this;

      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      _this = _React$Component.call.apply(_React$Component, [this].concat(args)) || this;

      _defineProperty(_assertThisInitialized(_this), "state", {
        error: null
      });

      return _this;
    }

    var _proto = ErrorBoundary.prototype;

    _proto.componentDidCatch = function componentDidCatch(error) {
      this.setState({
        error: error
      });
      errorCallback(error);
    };

    _proto.render = function render() {
      if (this.state.error) return null;
      return typeof Element === 'function' ? React.createElement(Element, null) : Element;
    };

    return ErrorBoundary;
  }(React.Component), _temp;
};

var generateElement = function generateElement(_ref, errorCallback) {
  var code = _ref.code,
      scope = _ref.scope,
      type = _ref.type;

  if (errorCallback === void 0) {
    errorCallback = function errorCallback() {};
  }

  try {
    var trimmedCode = code ? code.trim() : '';
    if (!trimmedCode) return null;
    var transformedCode = transform(prepareCode(trimmedCode), {
      transforms: ['jsx', 'imports', type === 'typescript' && 'typescript', type === 'flow' && 'flow'].filter(Boolean),
      production: true
    }).code.substr(13);
    var Element = errorBoundary(evalCode(transformedCode, _extends({
      React: React
    }, scope)), errorCallback);
    return React.createElement(Element, null);
  } catch (error) {
    // run after return
    setTimeout(errorCallback, 0, error);
    return null;
  }
};

var Runner =
/*#__PURE__*/
function (_React$Component) {
  _inheritsLoose(Runner, _React$Component);

  function Runner() {
    var _this;

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _React$Component.call.apply(_React$Component, [this].concat(args)) || this;

    _defineProperty(_assertThisInitialized(_this), "state", {
      error: null,
      element: null
    });

    _defineProperty(_assertThisInitialized(_this), "transpile", function () {
      var _this$props = _this.props,
          code = _this$props.code,
          scope = _this$props.scope,
          type = _this$props.type;
      var element = generateElement({
        code: code,
        scope: scope,
        type: type
      }, function (error) {
        _this.setState({
          error: error,
          element: null
        });
      });

      _this.setState({
        element: element,
        error: null
      });
    });

    return _this;
  }

  var _proto = Runner.prototype;

  _proto.componentDidMount = function componentDidMount() {
    this.transpile();
  };

  _proto.componentDidUpdate = function componentDidUpdate(prevProps) {
    var _this$props2 = this.props,
        code = _this$props2.code,
        scope = _this$props2.scope,
        type = _this$props2.type;

    if (code !== prevProps.code || scope !== prevProps.scope || type !== prevProps.type) {
      this.transpile();
    }
  };

  _proto.render = function render() {
    var children = this.props.children;
    var _this$state = this.state,
        element = _this$state.element,
        error = _this$state.error;
    return children({
      element: element,
      error: error
    });
  };

  return Runner;
}(React.Component);
Runner.propTypes = {
  children: PropTypes.func.isRequired,
  code: PropTypes.string.isRequired,
  scope: PropTypes.object,
  type: PropTypes.oneOf(['typescript', 'flow'])
};

var useRunner = function useRunner(_ref) {
  var code = _ref.code,
      scope = _ref.scope,
      type = _ref.type;
  if (!useState) throw new Error('Require React 16.8 or above to use hooks');

  var _useState = useState({
    element: null,
    error: null
  }),
      state = _useState[0],
      setState = _useState[1];

  useEffect(function () {
    var element = generateElement({
      code: code,
      scope: scope,
      type: type
    }, function (error) {
      setState({
        error: error,
        element: null
      });
    });
    setState({
      element: element,
      error: null
    });
  }, [code, scope, type]);
  return state;
};

export default Runner;
export { useRunner };
